/**
 * Copyright (C) 2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.medor.clone.lib;

import org.objectweb.medor.clone.api.Cloneable;

import java.util.Map;
import java.util.HashMap;

/**
 * is a basic implemation of the Cloneable interface. Each cloneable object
 * should extends this class and override the clone(Object,Map) method such as
 * the following example:<br/>
 * public class Toto extends BasicCloneable {
 *   String f1;
 *   Toto ref;
 *
 *   public Object clone(Object clone, Map obj2clone) throws CloneNotSupportedException {
 *     clone = super.clone(clone, obj2clone);
 *     ((Toto) clone).f1 = f1;
 *     ((Toto) clone).ref = (Toto) getClone(ref, obj2clone);
 *     return clone;
 *   }
 * }
 *
 * @author S.Chassande-Barrioz
 */
public class BasicCloneable implements Cloneable {
	public Object clone() throws CloneNotSupportedException {
		return clone(null, new HashMap());
	}

	/**
	 * Shortcut method permitting to clone the current instance.
	 * @param obj2clone is the cloning context.
	 * @return a clone of the current object (never null)
	 * @throws CloneNotSupportedException if the cloning is not possible due
	 * to a problem.
	 */
	public final Object clone(Map obj2clone) throws CloneNotSupportedException {
		if (obj2clone == null) {
			return this.clone(null, null);
		}
		Object o = obj2clone.get(this);
		if (o == null) {
			return this.clone(null, obj2clone);
		}
		return o;
	}

	/**
	 * Root implementation of the class tree.
	 * If the clone parameter is null, a clone of the current instance is
	 * instanciate throught getClass().newInstance() way.
	 * The clone instance (previoulsy instanciated or the parameter) is
	 * registered in the cloning context if the context is not null.
	 * @param clone is the clone instance to register. If this parameter is
	 * null, a new instance of the current object is instanciated
	 * (getClass().newInstance()). if this parameter is not null, it must be an
	 * instance of the current class.
	 * @param obj2clone is the cloning context used to register the clone. If
	 * this parameter is null, no registration is done.
	 * @return a clone of the current instance.
	 * @throws CloneNotSupportedException if the cloning is not possible due
	 * to a problem. A typically problem is when the clone parameter is null
	 * and when no no arg constructor is availlable in the class of the
	 * current instance.
	 */
	public Object clone(Object clone,
						Map obj2clone) throws CloneNotSupportedException {
		if (clone == null) {
			try {
				clone = getClass().newInstance();
			} catch (Exception e) {
				throw new CloneNotSupportedException("Impossible to clone a "
					+ getClass().getName() + " instance: " + e.getMessage());
			}
		} else if (clone.getClass().equals(getClass())) {
			throw new CloneNotSupportedException(
				"The clone parameter is not an instance of the current instance,"
				+ " current: " + getClass().getName()
				+ ", parameter class: " + clone.getClass().getName());
		}
		if (obj2clone != null) {
			obj2clone.put(this, clone);
		}
		return clone;
	}

	/**
	 * This static method is a shortcut method permitting to fetch the clone of
	 * an object. If the cloning context is not null, this implementation check
	 * if a clone is already available in the context. If yes the clone instance
	 * is returned. Otherwise, the cloning is launch on the obj parameter.
	 * @param obj is the object to clone. If this parameter is null, a null
	 * value is returned.
	 * @param obj2clone is the cloning context
	 * @return an clone instance of the obj parameter. The returned value is null
	 * only if the obj parameter is null.
	 * @throws CloneNotSupportedException if the cloning is not possible due
	 * to a problem.
	 */
	public static Object getClone(Cloneable obj,
						   Map obj2clone) throws CloneNotSupportedException {
		if (obj == null) {
			return null;
		}
		if (obj2clone == null) {
			return obj.clone(null, null);
		}
		Object o = obj2clone.get(obj);
		if (o == null) {
			return obj.clone(null, obj2clone);
		}
		return o;
	}
}
