/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.expression.parser.string;

import org.objectweb.jorm.type.api.PType;
import org.objectweb.medor.expression.api.Comparator;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ExpressionException;
import org.objectweb.medor.expression.api.MalformedExpressionException;
import org.objectweb.medor.expression.api.Operator;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.expression.lib.ConditionalAnd;
import org.objectweb.medor.expression.lib.ConditionalOr;
import org.objectweb.medor.expression.lib.DivideBy;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.Greater;
import org.objectweb.medor.expression.lib.GreaterEqual;
import org.objectweb.medor.expression.lib.Lower;
import org.objectweb.medor.expression.lib.LowerEqual;
import org.objectweb.medor.expression.lib.Minus;
import org.objectweb.medor.expression.lib.Mod;
import org.objectweb.medor.expression.lib.Mult;
import org.objectweb.medor.expression.lib.NotEqual;
import org.objectweb.medor.expression.lib.Or;
import org.objectweb.medor.expression.lib.Plus;
import org.objectweb.medor.expression.parser.api.ParameterTypeProvider;
import org.objectweb.medor.expression.parser.lib.ReplaceStringPlusByConcat;

import java.util.ArrayList;
import java.util.StringTokenizer;

/**
 * This class is a string parser of expressions.
 *
 * @author	P. Dechamboux, A.Lefebvre, S.Chassande-Barrioz
 */
public class ExpressionParser {
	/**
	 * The list of string constants extracted from the parsed expression.
	 */
	/**
	 * This token represents a string constant to be picked from the list of
	 * extracted ones.
	 */
	private final static String STRCONSTTOK = "$";
	private final static String DELIMITERS = "$()+-=!<>&|/%*";
	private final static int TOKENPOS = 0;
	private final static int STRCONSTPOS = 1;

	/**
	 * Produces a MEDOR Expression from a string.
	 *
	 * @param exprstr is the string expression to parse
	 * @param ptp is the parameter type provider. It permits to find the type
	 * of parameter found in the expression. If the expression does not contain
	 * parameter, no ParameterTypeProvider is required.
	 * @return the corresponding MEDOR Expression
	 * @throws ExpressionException
	 */
	public Expression parse(String exprstr, ParameterTypeProvider ptp)
			throws ExpressionException {
		ArrayList stringConst = new ArrayList();
		//Extract String constants and replace them by a $
		String exprwithoutstrconst = extractStringConst(exprstr, stringConst);
		//Parse
		Expression res = buildExpr(toStringArray(exprwithoutstrconst),
				new int[]{0, 0},
				stringConst,
				ptp);
		//replace ("toto" + "_titi") by concat("toto","_titi")
		res = new ReplaceStringPlusByConcat().rewrite(res);
		//Compute types
		res.compileExpression();
		return res;
	}

	/**
	 * Extracts all string constants from a expression string and stores them
	 * into the stringConst ArrayList, in the order they appear in the input
	 * expression. It also builds a new expression string which is the method
	 * result. All string constants has been replaced by a '$' character.
	 * @param expr	The input expression.
	 * @return		The resulting expression string.
	 * @throws MalformedExpressionException Thrown when a malformed string constant is
	 * found in the input expression string (non closed string constant).
	 */
	String extractStringConst(String expr, ArrayList stringConst)
			throws MalformedExpressionException {
		StringBuffer res = new StringBuffer();
		int curpos = 0;
		do {
			int startsc = expr.indexOf('"', curpos);
			if (startsc == -1) {
				break;
			}
			int endsc = expr.indexOf('"', startsc + 1);
			if (endsc == -1) {
				throw new MalformedExpressionException("Constant string starting at position"
											 + startsc + " never ended.");
			}
			res.append(expr.substring(curpos, startsc).replaceAll(" ", ""));
			res.append(STRCONSTTOK);
			stringConst.add(expr.substring(startsc + 1, endsc));
			curpos = endsc + 1;
		} while (true);
		res.append(expr.substring(curpos).replaceAll(" ", ""));
		return res.toString();
	}

	/**
	 *
	 * @param str
	 * @return
	 */
	private String[] toStringArray(String str) {
		StringTokenizer tokenizer = new StringTokenizer(str, DELIMITERS, true);
		ArrayList tokens = new ArrayList();
		String[] nextToken = new String[]{null};
		do {
			String token = getNextToken(tokenizer, nextToken);
			if (token == null) {
				break;
			}
			tokens.add(token);
		} while (true);
		return (String[]) tokens.toArray(new String[tokens.size()]);
	}

	/**
	 *
	 * @param tokens		The array of tokens to be parsed.
	 * @param indexes		The indexes of the current token to be parsed (pos 0)
	 * 						and of the next string constant to retrieve (pos 1).
	 * @param stringConst	The list of string constants.
	 * @return				The expression built by the method.
	 * @throws MalformedExpressionException
	 */
	private Expression buildExpr(String[] tokens,
								 int[] indexes,
								 ArrayList stringConst,
								 ParameterTypeProvider ptp)
			throws MalformedExpressionException {
		if (indexes[TOKENPOS] == tokens.length) {
			return null;
		}
		Expression firstnode;
		if (tokens[indexes[TOKENPOS]].equals("(")) {
			firstnode = buildParenthesExpr(tokens, indexes, stringConst, ptp);
		} else {
			firstnode = buildAtomicOperandExpr(tokens, indexes, stringConst, ptp);
		}
		if ((indexes[TOKENPOS] == tokens.length) || (tokens[indexes[TOKENPOS]].equals(")"))) {
			return firstnode;
		}
		Expression operator = buildNextExpr(tokens, indexes, stringConst, firstnode, ptp);
		if (operator == null) {
			throw new MalformedExpressionException("Wrong token found in token position "
										 + indexes[TOKENPOS] + ": " + tokens[indexes[TOKENPOS]]
										 + " (should be an operator)");
		}
		return operator;
	}

	/**
	 *
	 * @param tokens		The array of tokens to be parsed.
	 * @param indexes		The indexes of the current token to be parsed (pos 0)
	 * 						and of the next string constant to retrieve (pos 1).
	 * @param stringConst	The list of string constants.
	 * @return				The expression built by the method.
	 * @throws MalformedExpressionException
	 */
	private Operator buildNextExpr(String[] tokens,
							 int[] indexes,
							 ArrayList stringConst,
							 Expression left,
						     ParameterTypeProvider ptp)
			throws MalformedExpressionException {
		if (indexes[TOKENPOS] == tokens.length) {
			return null;
		}
		if (tokens[indexes[TOKENPOS]].equals(")")) {
			return null;
		}
		Operator operator;
		Expression secondnode;
		operator = buildPri1OperatorExpr(tokens, indexes);
		if (operator != null) {
			if (tokens[indexes[TOKENPOS]].equals("(")) {
				secondnode = buildParenthesExpr(tokens, indexes, stringConst, ptp);
			} else {
				secondnode = buildAtomicOperandExpr(tokens, indexes, stringConst, ptp);
			}
			Operator next = buildNextExpr(tokens, indexes, stringConst, operator, ptp);
			operator.setExpression(0, left);
			if (next != null) {
				if (priorTo(operator, next)) {
					operator.setExpression(1, secondnode);
					return next;
				}
				operator.setExpression(1, next);
				next.setExpression(0, secondnode);
				return operator;
			}
			operator.setExpression(1, secondnode);
			return operator;
		}
		operator = buildPri2OperatorExpr(tokens, indexes);
		if (operator != null) {
			if (tokens[indexes[TOKENPOS]].equals("(")) {
				secondnode = buildParenthesExpr(tokens, indexes, stringConst, ptp);
			} else {
				secondnode = buildAtomicOperandExpr(tokens, indexes, stringConst, ptp);
			}
			Operator next = buildNextExpr(tokens, indexes, stringConst, operator, ptp);
			operator.setExpression(0, left);
			if (next != null) {
				if (priorTo(operator, next)) {
					operator.setExpression(1, secondnode);
					return next;
				}
				operator.setExpression(1, next);
				next.setExpression(0, secondnode);
				return operator;
			}
			operator.setExpression(1, secondnode);
			return operator;
		}
		operator = buildCompOperatorExpr(tokens, indexes);
		if (operator != null) {
			if (tokens[indexes[TOKENPOS]].equals("(")) {
				secondnode = buildParenthesExpr(tokens, indexes, stringConst, ptp);
			} else {
				secondnode = buildExpr(tokens, indexes, stringConst, ptp);
			}
			Operator next = buildNextExpr(tokens, indexes, stringConst, operator, ptp);
			operator.setExpression(0, left);
			if (next != null) {
				if (priorTo(operator, next)) {
					operator.setExpression(1, secondnode);
					return next;
				}
				operator.setExpression(1, next);
				next.setExpression(0, secondnode);
				return operator;
			}
			operator.setExpression(1, secondnode);
			return operator;
		}
		operator = buildLogicalOperatorExpr(tokens, indexes);
		if (operator != null) {
			if (tokens[indexes[TOKENPOS]].equals("(")) {
				secondnode = buildParenthesExpr(tokens, indexes, stringConst, ptp);
			} else {
				secondnode = buildExpr(tokens, indexes, stringConst, ptp);
			}
			Operator next = buildNextExpr(tokens, indexes, stringConst, operator, ptp);
			operator.setExpression(0, left);
			if (next != null) {
				if (priorTo(operator, next)) {
					operator.setExpression(1, secondnode);
					return next;
				}
				operator.setExpression(1, next);
				next.setExpression(0, secondnode);
				return operator;
			}
			operator.setExpression(1, secondnode);
			return operator;
		}
		throw new MalformedExpressionException("Wrong token found in token position "
									 + indexes[TOKENPOS] + ": " + tokens[indexes[TOKENPOS]]
									 + " (should be an operator)");
	}

	/**
	 *
	 * @param tokens		The array of tokens to be parsed.
	 * @param indexes		The indexes of the current token to be parsed (pos 0)
	 * 						and of the next string constant to retrieve (pos 1).
	 * @param stringConst	The list of string constants.
	 * @return				The expression built by the method.
	 * @throws MalformedExpressionException
	 */
	private Expression buildParenthesExpr(String[] tokens,
								  int[] indexes,
								  ArrayList stringConst,
								  ParameterTypeProvider ptp)
			throws MalformedExpressionException {
		indexes[TOKENPOS]++;
		Expression res = buildExpr(tokens, indexes, stringConst, ptp);
		if (!tokens[indexes[TOKENPOS]].equals(")")) {
			throw new MalformedExpressionException("Wrong token found in token position "
										 + indexes[TOKENPOS] + ": " + tokens[indexes[TOKENPOS]]);
		}
		indexes[TOKENPOS]++;
		return res;
	}

	/**
	 * Looks for an atomic operand. It should be either a constant (string or
	 * numeric), or a variable.
	 * @param tokens		The array of tokens to be parsed.
	 * @param indexes		The indexes of the current token to be parsed (pos 0)
	 * 						and of the next string constant to retrieve (pos 1).
	 * @param stringConst	The list of string constants.
	 * @return				The expression built by the method.
	 * @throws MalformedExpressionException
	 */
	private Expression buildAtomicOperandExpr(String[] tokens,
									  int[] indexes,
									  ArrayList stringConst,
									  ParameterTypeProvider ptp)
			throws MalformedExpressionException {
		Expression res = null;
		String operand = tokens[indexes[TOKENPOS]];
		if (operand.equals("$")) { //constant string
			// String constant
			res = new BasicOperand(
					(String) stringConst.get(indexes[STRCONSTPOS]));
			indexes[STRCONSTPOS]++;
		}
		if (res == null) { // boolean
			if (operand.equalsIgnoreCase("true")) {
				res = new BasicOperand(true);
			} else if (operand.equalsIgnoreCase("false")) {
				res = new BasicOperand(false);
			}
		}
		if (res == null) { // integer
			try {
				long l = Long.parseLong(operand);
				if (l <= Byte.MAX_VALUE && l >= Byte.MIN_VALUE) {
					res = new BasicOperand((byte) l);
				} else if (l <= Short.MAX_VALUE && l >= Short.MIN_VALUE) {
					res = new BasicOperand((short) l);
				} else if (l <= Integer.MAX_VALUE && l >= Integer.MIN_VALUE) {
					res = new BasicOperand((int) l);
				} else {
					res = new BasicOperand(l);
				}
			} catch (NumberFormatException e) {
			}
		}
		if (res == null) { // floating
			try {
				double d = Double.parseDouble(operand);
				if (d <= Float.MAX_VALUE && d >= Float.MIN_VALUE) {
					res = new BasicOperand((float) d);
				} else {
					res = new BasicOperand(d);
				}
			} catch (NumberFormatException e) {
			}
		}
		if (res == null && Character.isLetterOrDigit(operand.charAt(0))) {
			// this is a variable expr to create
			String paramName = tokens[indexes[TOKENPOS]];
			if (ptp == null) {
				throw new MalformedExpressionException("ParameterTypeProvider required, but was null, for parameter " + paramName);
			}
			PType paramType = ptp.getParameterPType(paramName);
			if (paramType == null) {
				throw new MalformedExpressionException("Unknown type for parameter " + paramName);
			}
			res = new BasicParameterOperand(paramType, paramName);
		}
		if (res != null) {
			indexes[TOKENPOS]++;
			return res;
		}
		throw new MalformedExpressionException("Wrong token found in token position "
				+ indexes[TOKENPOS] + ": " + tokens[indexes[TOKENPOS]]
				+ " (should be a constant or a variable)");
	}

	/**
	 *
	 * @param tokens		The array of tokens to be parsed.
	 * @param indexes		The indexes of the current token to be parsed (pos 0)
	 * 						and of the next string constant to retrieve (pos 1).
	 * @return				The expression built by the method.
	 */
	private Operator buildLogicalOperatorExpr(String[] tokens,
										int[] indexes) {
		Operator res;
		if (tokens[indexes[TOKENPOS]].equals("&&")) {
			res = new ConditionalAnd();
		} else if (tokens[indexes[TOKENPOS]].equals("||")) {
			res = new ConditionalOr();
		} else {
			return null;
		}
		indexes[TOKENPOS]++;
		return res;
	}

	/**
	 *
	 * @param tokens		The array of tokens to be parsed.
	 * @param indexes		The indexes of the current token to be parsed (pos 0)
	 * 						and of the next string constant to retrieve (pos 1).
	 * @return				The expression built by the method.
	 */
	private Operator buildCompOperatorExpr(String[] tokens,
									 int[] indexes) {
		Operator res;
		if (tokens[indexes[TOKENPOS]].equals("==")) {
			res = new Equal();
		} else if (tokens[indexes[TOKENPOS]].equals("!=")) {
			res = new NotEqual();
		} else if (tokens[indexes[TOKENPOS]].equals("<")) {
			res = new Lower();
		} else if (tokens[indexes[TOKENPOS]].equals("<=")) {
			res = new LowerEqual();
		} else if (tokens[indexes[TOKENPOS]].equals(">")) {
			res = new Greater();
		} else if (tokens[indexes[TOKENPOS]].equals(">=")) {
			res = new GreaterEqual();
		} else {
			return null;
		}
		indexes[TOKENPOS]++;
		return res;
	}

	/**
	 *
	 * @param tokens		The array of tokens to be parsed.
	 * @param indexes		The indexes of the current token to be parsed (pos 0)
	 * 						and of the next string constant to retrieve (pos 1).
	 * @return				The expression built by the method.
	 */
	private Operator buildPri2OperatorExpr(String[] tokens,
									 int[] indexes) {
		Operator res;
		if (tokens[indexes[TOKENPOS]].equals("+")) {
			res = new Plus();
		} else if (tokens[indexes[TOKENPOS]].equals("-")) {
			res = new Minus();
		} else if (tokens[indexes[TOKENPOS]].equals("|")) {
			res = new Or();
		} else {
			return null;
		}
		indexes[TOKENPOS]++;
		return res;
	}

	/**
	 *
	 * @param tokens		The array of tokens to be parsed.
	 * @param indexes		The indexes of the current token to be parsed (pos 0)
	 * 						and of the next string constant to retrieve (pos 1).
	 * @return				The expression built by the method.
	 */
	private Operator buildPri1OperatorExpr(String[] tokens,
									 int[] indexes) {
		Operator res;
		if (tokens[indexes[TOKENPOS]].equals("*")) {
			res = new Mult();
		} else if (tokens[indexes[TOKENPOS]].equals("/")) {
			res = new DivideBy();
		} else if (tokens[indexes[TOKENPOS]].equals("%")) {
			res = new Mod();
		} else if (tokens[indexes[TOKENPOS]].equals("&")) {
			res = new And();
		} else {
			return null;
		}
		indexes[TOKENPOS]++;
		return res;
	}

	/**
	 *
	 * @param tokenizer
	 * @param nexttoken
	 * @return
	 */
	private String getNextToken(StringTokenizer tokenizer, String[] nexttoken) {
		String res;
		if (nexttoken[0] != null) {
			res = nexttoken[0];
			nexttoken[0] = null;
			return res;
		}
		if (!tokenizer.hasMoreTokens()) {
			return null;
		}
		res = tokenizer.nextToken();
		if (res.equals("=") || res.equals("<") || res.equals(">") || res.equals("!")) {
			if (!tokenizer.hasMoreTokens()) {
				return res;
			}
			nexttoken[0] = tokenizer.nextToken();
			if (nexttoken[0].equals("=")) {
				nexttoken[0] = null;
				return res + "=";
			}
			return res;
		} else if (res.equals("&")) {
			if (!tokenizer.hasMoreTokens()) {
				return res;
			}
			nexttoken[0] = tokenizer.nextToken();
			if (nexttoken[0].equals("&")) {
				nexttoken[0] = null;
				return "&&";
			}
			return res;
		} else if (res.equals("|")) {
			if (!tokenizer.hasMoreTokens()) {
				return res;
			}
			nexttoken[0] = tokenizer.nextToken();
			if (nexttoken[0].equals("|")) {
				nexttoken[0] = null;
				return "||";
			}
			return res;
		} else {
			return res;
		}
	}

	/**
	 *
	 * @param token
	 * @return
	 */
	private boolean isLog(Operator token) {
		return (token instanceof ConditionalAnd)
				|| (token instanceof ConditionalOr);
	}

	/**
	 *
	 * @param token
	 * @return
	 */
	private boolean isComp(Operator token) {
		return (token instanceof Comparator);
	}

	/**
	 *
	 * @param token
	 * @return
	 */
	private boolean isPri2(Operator token) {
		return (token instanceof Plus)
		|| (token instanceof Minus)
		|| (token instanceof Or);
	}

	/**
	 *
	 * @param tok1
	 * @param tok2
	 * @return
	 */
	private boolean priorTo(Operator tok1, Operator tok2) {
		if (isLog(tok1)) {
			return false;
		}
		if (isComp(tok1)) {
			return isLog(tok2);
		}
		if (isPri2(tok1)) {
			return isLog(tok2) || isComp(tok2);
		}
		return isLog(tok2) || isComp(tok2) || isPri2(tok2);
	}
}
