/**
 *  MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */

package org.objectweb.medor.optim.jorm;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.api.Expression;
import org.objectweb.medor.expression.api.ParameterOperand;
import org.objectweb.medor.expression.api.Operand;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.filter.jorm.lib.EncodePName;
import org.objectweb.medor.filter.jorm.lib.SinglePName;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.jorm.TestJormHelper;
import org.objectweb.medor.optim.lib.DropUselessNodeRule;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.query.api.CalculatedField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.lib.Project;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.medor.query.lib.SelectProject;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Collection;

public abstract class TestJormGoUpDecodeRule extends TestJormPDHelper {

	/**
	 * main method to launch the tests manually
	 */
	public static void main(String[] args) {
		TestRunner.run(suite());
	}

	/**
	 * This method creates a TestSuite object with the current tests
	 */
	public static Test suite() {
		return new TestSuite(TestJormGoUpDecodeRule.class);
	}

	public TestJormGoUpDecodeRule(String testName, String loggerName) {
		super(testName, loggerName);
	}

	public Example getExtent(Collection fileNames,
							 String className,
							 String pnameFieldName,
							 String[] fieldNames) {
		Example e = super.getExtent(fileNames, className, pnameFieldName, fieldNames);
		e.extent.setPMapper(mapper, PROJECT_NAME);
		return e;
	}

	Example eb;
	Example ec;
	Example ed;

	protected void setUp() {
		logger.log(BasicLevel.DEBUG, "Start of setUp");
		eb = getRewrittenB();
		ec = getRewrittenC();
		ed = getRewrittenD();
		logger.log(BasicLevel.DEBUG, "End of setUp");
	}

	protected void tearDown() {
	}

	public QueryTree getBCQueryTree(TestJormHelper tjh) {
		try {
			if (eb == null || ec == null)
				fail("All objects are not created by jorm"
					+ "\n-eb=" + eb
					+ "\n-ec=" + ec);
			QueryTreeField f1, f2;
			// Join B.y = C.x
			JoinProject j1 = new JoinProject("");
			Expression e = new Equal(
				new BasicFieldOperand(eb.qt.getTupleStructure().getField(eb.clazz.getFQName() + ".y")),
				new BasicFieldOperand(ec.qt.getTupleStructure().getField(ec.clazz.getFQName() + ".x"))
			);
			j1.setQueryFilter(e);
			f1 = (QueryTreeField) eb.qt.getTupleStructure().getField(eb.clazz.getFQName() + ".object");
			j1.addPropagatedField(f1.getName(), f1.getType(), new QueryTreeField[]{f1});
			f2 = (QueryTreeField) ec.qt.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			j1.addPropagatedField(f2.getName(), f2.getType(), new QueryTreeField[]{f2});
			return j1;
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}
		return null;
	}

	public QueryTree getBCDQueryTree(TestJormHelper tjh) {
		try {
			if (eb == null || ec == null || ed == null)
				fail("All objects are not created by jorm"
					+ "\n-eb=" + eb
					+ "\n-ec=" + ec
					+ "\n-ed=" + ed);
			QueryTreeField f1, f2;

			// Join B.y = C.x
			JoinProject j1 = new JoinProject("");
			Expression e = new Equal(
				new BasicFieldOperand(eb.qt.getTupleStructure().getField(eb.clazz.getFQName() + ".y")),
				new BasicFieldOperand(ec.qt.getTupleStructure().getField(ec.clazz.getFQName() + ".x"))
			);
			j1.setQueryFilter(e);
			f1 = (QueryTreeField) eb.qt.getTupleStructure().getField(eb.clazz.getFQName() + ".object");
			j1.addPropagatedField(f1.getName(), f1.getType(), new QueryTreeField[]{f1});
			f2 = (QueryTreeField) ec.qt.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			j1.addPropagatedField(f2.getName(), f2.getType(), new QueryTreeField[]{f2});

			// Join B.object = C.object
			JoinProject j2 = new JoinProject("");
			f1 = (QueryTreeField) j1.getTupleStructure().getField(eb.clazz.getFQName() + ".object");
			f2 = (QueryTreeField) ed.qt.getTupleStructure().getField(ed.clazz.getFQName() + ".object");
			e = new Equal(new BasicFieldOperand(f1), new BasicFieldOperand(f2));
			j2.setQueryFilter(e);
			f1 = (QueryTreeField) ed.qt.getTupleStructure().getField(ed.clazz.getFQName() + ".c");
			f2 = (QueryTreeField) j1.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			j2.addPropagatedField(f1.getName(), f1.getType(), new QueryTreeField[]{f1});
			j2.addPropagatedField(f2.getName(), f2.getType(), new QueryTreeField[]{f2});

				QueryTreePrinter.printQueryTree(j2, logger);
			return j2;

		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}
		return null;
	}

	public void testPnameEqualParam1() {
		logger.log(BasicLevel.DEBUG, "Start of testPnameEqualParam1");
		QueryTree j1 = getBCQueryTree(this);
		SelectProject qt = new SelectProject("");

		try {
			//filter
			QueryTreeField f1 = (QueryTreeField)
				j1.getTupleStructure().getField(eb.clazz.getFQName() + ".object");
			ParameterOperand po = new BasicParameterOperand(f1.getType(), "param1");
			qt.setQueryFilter(new Equal(new BasicFieldOperand(f1), po));
			//projection
			f1 = (QueryTreeField)
				j1.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			qt.addPropagatedField(f1.getName(), f1.getType(), new QueryTreeField[]{f1});
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}

		logger.log(BasicLevel.DEBUG, "JormLeafRewriter created");

		QueryTree rewritten = null;
		try {
			rewritten = new JormGoUpDecodeRule().rewrite(qt);
			rewritten = new DropUselessNodeRule().rewrite(rewritten);
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "BC1 rewritten");
		logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(rewritten, logger);

		logger.log(BasicLevel.DEBUG, "Checking of One");

		try {
			eb = getRewrittenB();
			ec = getRewrittenC();
			ed = getRewrittenD();
			//build the expected tree
			QueryLeaf qlb = (QueryLeaf) ((QueryNode) eb.qt).getChildren()[0];
			QueryLeaf qlc = (QueryLeaf) ((QueryNode) ec.qt).getChildren()[0];
			//Node 1
			Expression e = new Equal(
				new BasicFieldOperand(qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".y")),
				new BasicFieldOperand(qlc.getTupleStructure().getField(ec.clazz.getFQName() + ".x")));
			QueryNode qn1 = new JoinProject(e, "");
			QueryTreeField qtf = null;
			qtf = (QueryTreeField) qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".c2");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".c1");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qlc.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			//node 2
			Field c1 = qn1.getTupleStructure().getField(eb.clazz.getFQName() + ".c1");
			Field c2 = qn1.getTupleStructure().getField(eb.clazz.getFQName() + ".c2");
			String paramName = "param1";
			e = new And(
				new Equal(
					new BasicFieldOperand(c1),
					new EncodePName(
						c1.getType(),
						paramName,
						"f1")),
				new Equal(
					new BasicFieldOperand(c2),
					new EncodePName(
						c2.getType(),
						paramName,
						"f2"))
			);
			QueryNode qn2 = new SelectProject(e, "");
			qtf = (QueryTreeField) qn1.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			qn2.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			//Node 3
			QueryNode qn3 = new Project("");
			qtf = (QueryTreeField) qn2.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			CalculatedField decodeField = (CalculatedField)
				ec.qt.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			SinglePName spn = (SinglePName) decodeField.getExpression();
			qn3.addCalculatedField(
				decodeField.getName(),
				decodeField.getType(),
				new SinglePName(
					new BasicFieldOperand(qtf),
					(Operand) spn.getExpression(1),
					spn.getType()));

			logger.log(BasicLevel.DEBUG, "BC1 expected");
			logger.log(BasicLevel.DEBUG, "=========================================");

				QueryTreePrinter.printQueryTree(qn3, logger);
			TestQueryTreeHelper.equals("", qn3, rewritten, logger);
			logger.log(BasicLevel.DEBUG, "End of testPnameEqualParam1");
		}
		catch (MedorException me) {
			me.printStackTrace();
			fail(me.getMessage());
		}
	}

	public void testPnameEqualParam2() {
		logger.log(BasicLevel.DEBUG, "Start of testPnameEqualParam2");
		QueryTree j1 = getBCQueryTree(this);
		SelectProject qt = new SelectProject("");

		try {
			//filter
			QueryTreeField f1 = (QueryTreeField)
				j1.getTupleStructure().getField(eb.clazz.getFQName() + ".object");
			ParameterOperand po = new BasicParameterOperand(f1.getType(), "param1");
			qt.setQueryFilter(new Equal(po, new BasicFieldOperand(f1)));
			//projection
			f1 = (QueryTreeField)
				j1.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			qt.addPropagatedField(f1.getName(), f1.getType(), new QueryTreeField[]{f1});
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}

		logger.log(BasicLevel.DEBUG, "JormLeafRewriter created");

		QueryTree rewritten = null;
		try {
			rewritten = new JormGoUpDecodeRule().rewrite(qt);
			rewritten = new DropUselessNodeRule().rewrite(rewritten);
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "BC2 rewritten");
		logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(rewritten, logger);

		logger.log(BasicLevel.DEBUG, "Checking of One");

		try {
			eb = getRewrittenB();
			ec = getRewrittenC();
			ed = getRewrittenD();
			//build the expected tree
			QueryLeaf qlb = (QueryLeaf) ((QueryNode) eb.qt).getChildren()[0];
			QueryLeaf qlc = (QueryLeaf) ((QueryNode) ec.qt).getChildren()[0];
			//Node 1
			Expression e = new Equal(
				new BasicFieldOperand(qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".y")),
				new BasicFieldOperand(qlc.getTupleStructure().getField(ec.clazz.getFQName() + ".x")));
			QueryNode qn1 = new JoinProject(e, "");
			QueryTreeField qtf = null;
			qtf = (QueryTreeField) qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".c2");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".c1");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qlc.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			//node 2
			Field c1 = qn1.getTupleStructure().getField(eb.clazz.getFQName() + ".c1");
			Field c2 = qn1.getTupleStructure().getField(eb.clazz.getFQName() + ".c2");
			String paramName = "param1";
			e = new And(
				new Equal(
					new EncodePName(
						c1.getType(),
						paramName,
						"f1"),
					new BasicFieldOperand(c1)
				),
				new Equal(
					new EncodePName(
						c2.getType(),
						paramName,
						"f2"),
					new BasicFieldOperand(c2)
				)
			);
			QueryNode qn2 = new SelectProject(e, "");
			qtf = (QueryTreeField) qn1.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			qn2.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			//Node 3
			QueryNode qn3 = new Project("");
			qtf = (QueryTreeField) qn2.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			CalculatedField decodeField = (CalculatedField)
				ec.qt.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			SinglePName spn = (SinglePName) decodeField.getExpression();
			qn3.addCalculatedField(
				decodeField.getName(),
				decodeField.getType(),
				new SinglePName(
					new BasicFieldOperand(qtf),
					(Operand) spn.getExpression(1),
					spn.getType()));

			logger.log(BasicLevel.DEBUG, "BC2 expected");
			logger.log(BasicLevel.DEBUG, "=========================================");

				QueryTreePrinter.printQueryTree(qn3, logger);
			TestQueryTreeHelper.equals("", qn3, rewritten, logger);
		}
		catch (MedorException me) {
			me.printStackTrace();
			fail(me.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "End of testPnameEqualParam2");
	}

	public void testPnameEqualParam3() {
		logger.log(BasicLevel.DEBUG, "Start of testPnameEqualParam3");
		QueryTree j1 = getBCQueryTree(this);
		SelectProject qt = new SelectProject("");

		try {
			//filter
			QueryTreeField f1 = (QueryTreeField)
				j1.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			ParameterOperand po = new BasicParameterOperand(f1.getType(), "param1");
			qt.setQueryFilter(new Equal(po, new BasicFieldOperand(f1)));
			//projection
			f1 = (QueryTreeField)
				j1.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			qt.addPropagatedField(f1.getName(), f1.getType(), new QueryTreeField[]{f1});
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}

		logger.log(BasicLevel.DEBUG, "JormLeafRewriter created");

		QueryTree rewritten = null;
		try {
			rewritten = new JormGoUpDecodeRule().rewrite(qt);

			rewritten = new DropUselessNodeRule().rewrite(rewritten);
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "BC3 rewritten");
		logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(rewritten, logger);

		logger.log(BasicLevel.DEBUG, "Checking of One");

		try {
			eb = getRewrittenB();
			ec = getRewrittenC();
			ed = getRewrittenD();
			//build the expected tree
			QueryLeaf qlb = (QueryLeaf) ((QueryNode) eb.qt).getChildren()[0];
			QueryLeaf qlc = (QueryLeaf) ((QueryNode) ec.qt).getChildren()[0];
			//Node 1
			Expression e = new Equal(
				new BasicFieldOperand(qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".y")),
				new BasicFieldOperand(qlc.getTupleStructure().getField(ec.clazz.getFQName() + ".x")));
			QueryNode qn1 = new JoinProject(e, "");
			QueryTreeField qtf = null;
			qtf = (QueryTreeField) qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".c2");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".c1");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qlc.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			//node 2
			Field id = qn1.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			String paramName = "param1";
			e = new Equal(
				new EncodePName(id.getType(), paramName),
				new BasicFieldOperand(id));
			QueryNode qn2 = new SelectProject(e, "");
			qtf = (QueryTreeField) qn1.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			qn2.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			//Node 3
			QueryNode qn3 = new Project("");
			qtf = (QueryTreeField) qn2.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			CalculatedField decodeField = (CalculatedField)
				ec.qt.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			SinglePName spn = (SinglePName) decodeField.getExpression();
			qn3.addCalculatedField(
				decodeField.getName(),
				decodeField.getType(),
				new SinglePName(
					new BasicFieldOperand(qtf),
					(Operand) spn.getExpression(1),
					spn.getType()));

			logger.log(BasicLevel.DEBUG, "BC3 expected");
			logger.log(BasicLevel.DEBUG, "=========================================");

				QueryTreePrinter.printQueryTree(qn3, logger);
			TestQueryTreeHelper.equals("", qn3, rewritten, logger);
		}
		catch (MedorException me) {
			me.printStackTrace();
			fail(me.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "End of testPnameEqualParam3");
	}

	public void testBCD() {
		logger.log(BasicLevel.DEBUG, "Start of testBCD");
		QueryTree qt = getBCDQueryTree(this);
		logger.log(BasicLevel.DEBUG, "JormLeafRewriter created");

		QueryTree rewritten = null;
		try {
			rewritten = new JormGoUpDecodeRule().rewrite(qt);
			logger.log(BasicLevel.DEBUG, "--------------After GoUpDecode: ");
			QueryTreePrinter.printQueryTree(rewritten, logger);
			logger.log(BasicLevel.DEBUG, "--------------Entering DropUselessNode ");
			
			rewritten = new DropUselessNodeRule().rewrite(rewritten);
		}
		catch (MedorException e) {
			e.printStackTrace();
			fail(e.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "BCD rewritten");
		logger.log(BasicLevel.DEBUG, "=========================================");

			QueryTreePrinter.printQueryTree(rewritten, logger);

		logger.log(BasicLevel.DEBUG, "Checking of One");

		try {
			//build the expected tree
			Example eb = getRewrittenB();
			QueryLeaf qlb = (QueryLeaf) ((QueryNode) eb.qt).getChildren()[0];

			Example ec = getRewrittenC();
			QueryLeaf qlc = (QueryLeaf) ((QueryNode) ec.qt).getChildren()[0];

			Example ed = getRewrittenD();
			QueryLeaf qld = (QueryLeaf) ((QueryNode) ed.qt).getChildren()[0];

			Expression e = new Equal(
				new BasicFieldOperand(qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".y")),
				new BasicFieldOperand(qlc.getTupleStructure().getField(ec.clazz.getFQName() + ".x")));
			QueryNode qn1 = new JoinProject(e, "");
			QueryTreeField qtf = null;
			qtf = (QueryTreeField) qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".c2");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qlb.getTupleStructure().getField(eb.clazz.getFQName() + ".c1");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qlc.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			qn1.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			e = new And(
				new Equal(
					new BasicFieldOperand(qn1.getTupleStructure().getField(eb.clazz.getFQName() + ".c1")),
					new BasicFieldOperand(qld.getTupleStructure().getField(ed.clazz.getFQName() + ".a"))
				),
				new Equal(
					new BasicFieldOperand(qn1.getTupleStructure().getField(eb.clazz.getFQName() + ".c2")),
					new BasicFieldOperand(qld.getTupleStructure().getField(ed.clazz.getFQName() + ".b"))
				)
			);
			QueryNode qn2 = new JoinProject(e, "");
			qtf = (QueryTreeField) qld.getTupleStructure().getField(ed.clazz.getFQName() + ".c");
			qn2.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});
			qtf = (QueryTreeField) qn1.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			qn2.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			QueryNode qn3 = new Project("");
			qtf = (QueryTreeField) qn2.getTupleStructure().getField(ec.clazz.getFQName() + ".id");
			CalculatedField decodeField = (CalculatedField)
				ec.qt.getTupleStructure().getField(ec.clazz.getFQName() + ".object");
			SinglePName spn = (SinglePName) decodeField.getExpression();
			qn3.addCalculatedField(
				decodeField.getName(),
				decodeField.getType(),
				new SinglePName(
					new BasicFieldOperand(qtf),
					(Operand) spn.getExpression(1),
					spn.getType()));
			qtf = (QueryTreeField) qn2.getTupleStructure().getField(ed.clazz.getFQName() + ".c");
			qn3.addPropagatedField(qtf.getName(), qtf.getType(), new QueryTreeField[]{qtf});

			TestQueryTreeHelper.equals("", qn3, rewritten, logger);
		}
		catch (MedorException me) {
			me.printStackTrace();
			fail(me.getMessage());
		}
		logger.log(BasicLevel.DEBUG, "End of testBCD");
	}

	public void testMemberOf() {
		logger.log(BasicLevel.DEBUG, "Start of testMemberOf");
		Example e1 = getRewrittenOne();
		Example ea = getRewrittenA();
		try {
			QueryTreeField qfa =  (QueryTreeField)
				e1.qt.getTupleStructure().getField(
					e1.clazz.getFQName() + ".fieldA");
			ArrayList al1 = new ArrayList(1);
			al1.add(new BasicFieldOperand(qfa));
			ArrayList al2 = new ArrayList(1);
			al2.add(new BasicFieldOperand(ea.qt.getTupleStructure().getField(
				ea.clazz.getFQName() + ".object")));
			Expression e = new MemberOf(al1, al2);
			QueryNode qn2 = new JoinProject(e, "");
			qn2.addPropagatedField("toto", qfa.getType(), new QueryTreeField[]{qfa});
			logger.log(BasicLevel.DEBUG, "QueryTree before rewriting");
			QueryTreePrinter.printQueryTree(qn2, logger);

			QueryTree rewritten = new JormGoUpDecodeRule().rewrite(qn2);
			logger.log(BasicLevel.DEBUG, "QueryTree after rewriting");
            QueryTreePrinter.printQueryTree(rewritten, logger);
 		}
		catch (Exception me) {
			me.printStackTrace();
			fail(me.getMessage());

		}
		logger.log(BasicLevel.DEBUG, "End of testMemberOf");

	}
}
